@@ -54,7 +54,7 @@ static char * wpa_config_get_line(char *s, int size, FILE *stream, int *line)
 		 * string. Remove trailing white space. */
 		sstart = strchr(pos, '"');
 		if (sstart)
-			sstart = strchr(sstart + 1, '"');
+			sstart = strrchr(sstart + 1, '"');
 		if (!sstart)
 			sstart = pos;
 		end = strchr(sstart, '#');
@@ -119,7 +119,8 @@ static int wpa_config_parse_str(struct parse_data *data,
 	*dst = wpa_config_parse_string(value, &res_len);
 	if (*dst == NULL) {
 		wpa_printf(MSG_ERROR, "Line %d: failed to parse %s '%s'.",
-			   line, data->name, value);
+			   line, data->name,
+			   data->key_data ? "[KEY DATA REMOVED]" : value);
 		return -1;
 	}
 	if (data->param2)
@@ -508,6 +509,7 @@ static int wpa_config_parse_eap(struct parse_data *data, int line,
 		methods = realloc(methods, num_methods + 1);
 		if (methods == NULL) {
 			free(tmp);
+			free(buf);
 			return -1;
 		}
 		methods[num_methods] = eap_get_type(start);
