@@ -148,7 +148,7 @@ static int wpa_supplicant_ctrl_iface_attach(struct wpa_supplicant *wpa_s,
 		   inet_ntoa(from->sin_addr), ntohs(from->sin_port));
 #else /* CONFIG_CTRL_IFACE_UDP */
 	wpa_hexdump(MSG_DEBUG, "CTRL_IFACE monitor attached",
-		    (u8 *) from->sun_path, fromlen);
+		    (u8 *) from->sun_path, fromlen - sizeof(from->sun_family));
 #endif /* CONFIG_CTRL_IFACE_UDP */
 	return 0;
 }
@@ -177,14 +177,16 @@ static int wpa_supplicant_ctrl_iface_detach(struct wpa_supplicant *wpa_s,
 		}
 #else /* CONFIG_CTRL_IFACE_UDP */
 		if (fromlen == dst->addrlen &&
-		    memcmp(from->sun_path, dst->addr.sun_path, fromlen) == 0) {
+		    memcmp(from->sun_path, dst->addr.sun_path,
+			   fromlen - sizeof(from->sun_family)) == 0) {
 			if (prev == NULL)
 				wpa_s->ctrl_dst = dst->next;
 			else
 				prev->next = dst->next;
 			free(dst);
 			wpa_hexdump(MSG_DEBUG, "CTRL_IFACE monitor detached",
-				    (u8 *) from->sun_path, fromlen);
+				    (u8 *) from->sun_path,
+				    fromlen - sizeof(from->sun_family));
 			return 0;
 		}
 #endif /* CONFIG_CTRL_IFACE_UDP */
@@ -217,9 +219,11 @@ static int wpa_supplicant_ctrl_iface_level(struct wpa_supplicant *wpa_s,
 		}
 #else /* CONFIG_CTRL_IFACE_UDP */
 		if (fromlen == dst->addrlen &&
-		    memcmp(from->sun_path, dst->addr.sun_path, fromlen) == 0) {
+		    memcmp(from->sun_path, dst->addr.sun_path,
+			   fromlen - sizeof(from->sun_family)) == 0) {
 			wpa_hexdump(MSG_DEBUG, "CTRL_IFACE changed monitor "
-				    "level", (u8 *) from->sun_path, fromlen);
+				    "level", (u8 *) from->sun_path,
+				    fromlen - sizeof(from->sun_family));
 			dst->debug_level = atoi(level);
 			return 0;
 		}
@@ -716,7 +720,8 @@ void wpa_supplicant_ctrl_iface_send(struct wpa_supplicant *wpa_s, int level,
 		next = dst->next;
 		if (level >= dst->debug_level) {
 			wpa_hexdump(MSG_DEBUG, "CTRL_IFACE monitor send",
-				    (u8 *) dst->addr.sun_path, dst->addrlen);
+				    (u8 *) dst->addr.sun_path, dst->addrlen -
+				    sizeof(dst->addr.sun_family));
 			msg.msg_name = &dst->addr;
 			msg.msg_namelen = dst->addrlen;
 			if (sendmsg(wpa_s->ctrl_sock, &msg, 0) < 0) {
