@@ -36,9 +36,20 @@ struct wpa_driver_wext_data {
 	int event_sock;
 	int ioctl_sock;
 	char ifname[IFNAMSIZ + 1];
+	u8 *assoc_req_ies;
+	size_t assoc_req_ies_len;
+	u8 *assoc_resp_ies;
+	size_t assoc_resp_ies_len;
+	struct wpa_driver_capa capa;
+	int has_capability;
+	int we_version_compiled;
 };
 
 
+static int wpa_driver_wext_flush_pmkid(void *priv);
+static int wpa_driver_wext_get_range(void *priv);
+
+
 static int wpa_driver_wext_set_auth_param(struct wpa_driver_wext_data *drv,
 					  int index, u32 value)
 {
@@ -52,6 +63,8 @@ static int wpa_driver_wext_set_auth_param(struct wpa_driver_wext_data *drv,
 
 	if (ioctl(drv->ioctl_sock, SIOCSIWAUTH, &iwr) < 0) {
 		perror("ioctl[SIOCSIWAUTH]");
+		fprintf(stderr, "WEXT auth param %d value 0x%x - ",
+			index, value);
 		ret = -1;
 	}
 
@@ -115,8 +128,18 @@ int wpa_driver_wext_get_ssid(void *priv, u8 *ssid)
 	if (ioctl(drv->ioctl_sock, SIOCGIWESSID, &iwr) < 0) {
 		perror("ioctl[SIOCGIWESSID]");
 		ret = -1;
-	} else
+	} else {
 		ret = iwr.u.essid.length;
+		if (ret > 32)
+			ret = 32;
+		/* Some drivers include nul termination in the SSID, so let's
+		 * remove it here before further processing. WE-21 changes this
+		 * to explicitly require the length _not_ to include nul
+		 * termination. */
+		if (ret > 0 && ssid[ret - 1] == '\0' &&
+		    drv->we_version_compiled < 21)
+			ret--;
+	}
 
 	return ret;
 }
@@ -139,12 +162,19 @@ int wpa_driver_wext_set_ssid(void *priv, const u8 *ssid, size_t ssid_len)
 	memset(buf, 0, sizeof(buf));
 	memcpy(buf, ssid, ssid_len);
 	iwr.u.essid.pointer = (caddr_t) buf;
-	/* For historic reasons, set SSID length to include one extra
-	 * character, C string nul termination, even though SSID is really an
-	 * octet string that should not be presented as a C string. Some Linux
-	 * drivers decrement the length by one and can thus end up missing the
-	 * last octet of the SSID if the length is not incremented here. */
-	iwr.u.essid.length = ssid_len ? ssid_len + 1 : 0;
+	if (drv->we_version_compiled < 21) {
+		/* For historic reasons, set SSID length to include one extra
+		 * character, C string nul termination, even though SSID is
+		 * really an octet string that should not be presented as a C
+		 * string. Some Linux drivers decrement the length by one and
+		 * can thus end up missing the last octet of the SSID if the
+		 * length is not incremented here. WE-21 changes this to
+		 * explicitly require the length _not_ to include nul
+		 * termination. */
+		if (ssid_len)
+			ssid_len++;
+	}
+	iwr.u.essid.length = ssid_len;
 
 	if (ioctl(drv->ioctl_sock, SIOCSIWESSID, &iwr) < 0) {
 		perror("ioctl[SIOCSIWESSID]");
@@ -244,7 +274,7 @@ static int wpa_driver_wext_event_wireless_michaelmicfailure(
 	const struct iw_michaelmicfailure *mic;
 	union wpa_event_data data;
 
-	if (len < sizeof(mic))
+	if (len < sizeof(*mic))
 		return -1;
 
 	mic = (const struct iw_michaelmicfailure *) ev;
@@ -261,6 +291,101 @@ static int wpa_driver_wext_event_wireless_michaelmicfailure(
 }
 
 
+static int wpa_driver_wext_event_wireless_pmkidcand(
+	struct wpa_driver_wext_data *drv, void *ctx, const char *ev, int len)
+{
+	const struct iw_pmkid_cand *cand;
+	union wpa_event_data data;
+	const u8 *addr;
+
+	if (len < sizeof(*cand))
+		return -1;
+
+	cand = (const struct iw_pmkid_cand *) ev;
+	addr = (const u8 *) cand->bssid.sa_data;
+
+	wpa_printf(MSG_DEBUG, "PMKID candidate wireless event: "
+		   "flags=0x%x index=%d bssid=" MACSTR, cand->flags,
+		   cand->index, MAC2STR(addr));
+
+	memset(&data, 0, sizeof(data));
+	memcpy(data.pmkid_candidate.bssid, addr, ETH_ALEN);
+	data.pmkid_candidate.index = cand->index;
+	data.pmkid_candidate.preauth = cand->flags & IW_PMKID_CAND_PREAUTH;
+	wpa_supplicant_event(drv->ctx, EVENT_PMKID_CANDIDATE, &data);
+
+	return 0;
+}
+
+
+static int wpa_driver_wext_event_wireless_assocreqie(
+	struct wpa_driver_wext_data *drv, void *ctx, const char *ev, int len)
+{
+	if (len < 0)
+		return -1;
+
+	wpa_hexdump(MSG_DEBUG, "AssocReq IE wireless event", (const u8 *) ev,
+		    len);
+	free(drv->assoc_req_ies);
+	drv->assoc_req_ies = malloc(len);
+	if (drv->assoc_req_ies == NULL) {
+		drv->assoc_req_ies_len = 0;
+		return -1;
+	}
+	memcpy(drv->assoc_req_ies, ev, len);
+	drv->assoc_req_ies_len = len;
+
+	return 0;
+}
+
+
+static int wpa_driver_wext_event_wireless_assocrespie(
+	struct wpa_driver_wext_data *drv, void *ctx, const char *ev, int len)
+{
+	if (len < 0)
+		return -1;
+
+	wpa_hexdump(MSG_DEBUG, "AssocResp IE wireless event", (const u8 *) ev,
+		    len);
+	free(drv->assoc_resp_ies);
+	drv->assoc_resp_ies = malloc(len);
+	if (drv->assoc_resp_ies == NULL) {
+		drv->assoc_resp_ies_len = 0;
+		return -1;
+	}
+	memcpy(drv->assoc_resp_ies, ev, len);
+	drv->assoc_resp_ies_len = len;
+
+	return 0;
+}
+
+
+static void wpa_driver_wext_event_assoc_ies(struct wpa_driver_wext_data *drv)
+{
+	union wpa_event_data data;
+
+	if (drv->assoc_req_ies == NULL && drv->assoc_resp_ies == NULL)
+		return;
+
+	memset(&data, 0, sizeof(data));
+	if (drv->assoc_req_ies) {
+		data.assoc_info.req_ies = drv->assoc_req_ies;
+		drv->assoc_req_ies = NULL;
+		data.assoc_info.req_ies_len = drv->assoc_req_ies_len;
+	}
+	if (drv->assoc_resp_ies) {
+		data.assoc_info.resp_ies = drv->assoc_resp_ies;
+		drv->assoc_resp_ies = NULL;
+		data.assoc_info.resp_ies_len = drv->assoc_resp_ies_len;
+	}
+
+	wpa_supplicant_event(drv->ctx, EVENT_ASSOCINFO, &data);
+
+	free(data.assoc_info.req_ies);
+	free(data.assoc_info.resp_ies);
+}
+
+
 static void wpa_driver_wext_event_wireless(struct wpa_driver_wext_data *drv,
 					   void *ctx, char *data, int len)
 {
@@ -273,11 +398,29 @@ static void wpa_driver_wext_event_wireless(struct wpa_driver_wext_data *drv,
 	while (pos + IW_EV_LCP_LEN <= end) {
 		/* Event data may be unaligned, so make a local, aligned copy
 		 * before processing. */
-		memcpy(&iwe_buf, pos, sizeof(struct iw_event));
+		memcpy(&iwe_buf, pos, IW_EV_LCP_LEN);
 		wpa_printf(MSG_DEBUG, "Wireless event: cmd=0x%x len=%d",
 			   iwe->cmd, iwe->len);
 		if (iwe->len <= IW_EV_LCP_LEN)
 			return;
+
+		custom = pos + IW_EV_POINT_LEN;
+		if (drv->we_version_compiled > 18 &&
+		    (iwe->cmd == IWEVMICHAELMICFAILURE ||
+		     iwe->cmd == IWEVCUSTOM ||
+		     iwe->cmd == IWEVASSOCREQIE ||
+		     iwe->cmd == IWEVASSOCRESPIE ||
+		     iwe->cmd == IWEVPMKIDCAND)) {
+			/* WE-19 removed the pointer from struct iw_point */
+			char *dpos = (char *) &iwe_buf.u.data.length;
+			int dlen = dpos - (char *) &iwe_buf;
+			memcpy(dpos, pos + IW_EV_LCP_LEN,
+			       sizeof(struct iw_event) - dlen);
+		} else {
+			memcpy(&iwe_buf, pos, sizeof(struct iw_event));
+			custom += IW_EV_POINT_OFF;
+		}
+
 		switch (iwe->cmd) {
 		case SIOCGIWAP:
 			wpa_printf(MSG_DEBUG, "Wireless event: new AP: "
@@ -288,18 +431,24 @@ static void wpa_driver_wext_event_wireless(struct wpa_driver_wext_data *drv,
 			    ||
 			    memcmp(iwe->u.ap_addr.sa_data,
 				   "\x44\x44\x44\x44\x44\x44", ETH_ALEN) == 0)
+			{
+				free(drv->assoc_req_ies);
+				drv->assoc_req_ies = NULL;
+				free(drv->assoc_resp_ies);
+				drv->assoc_resp_ies = NULL;
 				wpa_supplicant_event(ctx, EVENT_DISASSOC,
 						     NULL);
-			else
+			
+			} else {
+				wpa_driver_wext_event_assoc_ies(drv);
 				wpa_supplicant_event(ctx, EVENT_ASSOC, NULL);
+			}
 			break;
 		case IWEVMICHAELMICFAILURE:
 			wpa_driver_wext_event_wireless_michaelmicfailure(
-				drv, ctx, pos + IW_EV_POINT_LEN,
-				iwe->u.data.length);
+				drv, ctx, custom, iwe->u.data.length);
 			break;
 		case IWEVCUSTOM:
-			custom = pos + IW_EV_POINT_LEN;
 			if (custom + iwe->u.data.length > end)
 				return;
 			buf = malloc(iwe->u.data.length + 1);
@@ -315,6 +464,18 @@ static void wpa_driver_wext_event_wireless(struct wpa_driver_wext_data *drv,
 					     drv, ctx);
 			wpa_supplicant_event(ctx, EVENT_SCAN_RESULTS, NULL);
 			break;
+		case IWEVASSOCREQIE:
+			wpa_driver_wext_event_wireless_assocreqie(
+				drv, ctx, custom, iwe->u.data.length);
+			break;
+		case IWEVASSOCRESPIE:
+			wpa_driver_wext_event_wireless_assocrespie(
+				drv, ctx, custom, iwe->u.data.length);
+			break;
+		case IWEVPMKIDCAND:
+			wpa_driver_wext_event_wireless_pmkidcand(
+				drv, ctx, custom, iwe->u.data.length);
+			break;
 		}
 
 		pos += iwe->len;
@@ -542,6 +703,11 @@ void * wpa_driver_wext_init(void *ctx, const char *ifname)
 	eloop_register_read_sock(s, wpa_driver_wext_event_receive, drv, ctx);
 	drv->event_sock = s;
 
+	/*
+	 * Make sure that the driver does not have any obsolete PMKID entries.
+	 */
+	wpa_driver_wext_flush_pmkid(drv);
+
 	if (wpa_driver_wext_set_mode(drv, 0) < 0) {
 		printf("Could not configure driver to use managed mode\n");
 	}
@@ -551,6 +717,8 @@ void * wpa_driver_wext_init(void *ctx, const char *ifname)
 		printf("Could not set interface '%s' UP\n", drv->ifname);
 	}
 
+	wpa_driver_wext_get_range(drv);
+
 	return drv;
 }
 
@@ -564,7 +732,7 @@ void wpa_driver_wext_deinit(void *priv)
 	 * Clear possibly configured driver parameters in order to make it
 	 * easier to use the driver after wpa_supplicant has been terminated.
 	 */
-	wpa_driver_wext_set_bssid(drv, "\x00\x00\x00\x00\x00\x00");
+	wpa_driver_wext_set_bssid(drv, (u8 *) "\x00\x00\x00\x00\x00\x00");
 
 	eloop_unregister_read_sock(drv->event_sock);
 
@@ -573,6 +741,8 @@ void wpa_driver_wext_deinit(void *priv)
 
 	close(drv->event_sock);
 	close(drv->ioctl_sock);
+	free(drv->assoc_req_ies);
+	free(drv->assoc_resp_ies);
 	free(drv);
 }
 
@@ -590,7 +760,6 @@ int wpa_driver_wext_scan(void *priv, const u8 *ssid, size_t ssid_len)
 	struct iwreq iwr;
 	int ret = 0;
 	struct iw_scan_req req;
-	int num = SIOCSIWSCAN;
 
 	if (ssid_len > IW_ESSID_MAX_SIZE) {
 		wpa_printf(MSG_DEBUG, "%s: too long SSID (%lu)",
@@ -602,7 +771,6 @@ int wpa_driver_wext_scan(void *priv, const u8 *ssid, size_t ssid_len)
 	strncpy(iwr.ifr_name, drv->ifname, IFNAMSIZ);
 
 	if (ssid && ssid_len) {
-		num = SIOCSIWSCANEXT;
 		memset(&req, 0, sizeof(req));
 		req.essid_len = ssid_len;
 		req.bssid.sa_family = ARPHRD_ETHER;
@@ -613,8 +781,8 @@ int wpa_driver_wext_scan(void *priv, const u8 *ssid, size_t ssid_len)
 		iwr.u.data.flags = IW_SCAN_THIS_ESSID;
 	}
 
-	if (ioctl(drv->ioctl_sock, num, &iwr) < 0) {
-		perror("ioctl[SIOCSIWSCAN{,EXT}]");
+	if (ioctl(drv->ioctl_sock, SIOCSIWSCAN, &iwr) < 0) {
+		perror("ioctl[SIOCSIWSCAN]");
 		ret = -1;
 	}
 
@@ -627,7 +795,7 @@ int wpa_driver_wext_scan(void *priv, const u8 *ssid, size_t ssid_len)
 }
 
 
-/* Compare function for sorting scan results. Return >0 if @b is consider
+/* Compare function for sorting scan results. Return >0 if @b is considered
  * better. */
 static int wpa_scan_result_compar(const void *a, const void *b)
 {
@@ -643,11 +811,11 @@ static int wpa_scan_result_compar(const void *a, const void *b)
 		return -1;
 
 	/* privacy support preferred */
-	if ((wa->caps & IW_ENCODE_DISABLED) &&
-	    (wb->caps & IW_ENCODE_DISABLED) == 0)
+	if ((wa->caps & IEEE80211_CAP_PRIVACY) == 0 &&
+	    (wb->caps & IEEE80211_CAP_PRIVACY))
 		return 1;
-	if ((wa->caps & IW_ENCODE_DISABLED) == 0 &&
-	    (wb->caps & IW_ENCODE_DISABLED))
+	if ((wa->caps & IEEE80211_CAP_PRIVACY) &&
+	    (wb->caps & IEEE80211_CAP_PRIVACY) == 0)
 		return -1;
 
 	/* best/max rate preferred if signal level close enough XXX */
@@ -714,9 +882,26 @@ int wpa_driver_wext_get_scan_results(void *priv,
 		int ssid_len;
 		/* Event data may be unaligned, so make a local, aligned copy
 		 * before processing. */
-		memcpy(&iwe_buf, pos, sizeof(struct iw_event));
+		memcpy(&iwe_buf, pos, IW_EV_LCP_LEN);
 		if (iwe->len <= IW_EV_LCP_LEN)
 			break;
+
+		custom = pos + IW_EV_POINT_LEN;
+		if (drv->we_version_compiled > 18 &&
+		    (iwe->cmd == SIOCGIWESSID ||
+		     iwe->cmd == SIOCGIWENCODE ||
+		     iwe->cmd == IWEVGENIE ||
+		     iwe->cmd == IWEVCUSTOM)) {
+			/* WE-19 removed the pointer from struct iw_point */
+			char *dpos = (char *) &iwe_buf.u.data.length;
+			int dlen = dpos - (char *) &iwe_buf;
+			memcpy(dpos, pos + IW_EV_LCP_LEN,
+			       sizeof(struct iw_event) - dlen);
+		} else {
+			memcpy(&iwe_buf, pos, sizeof(struct iw_event));
+			custom += IW_EV_POINT_OFF;
+		}
+
 		switch (iwe->cmd) {
 		case SIOCGIWAP:
 			if (!first)
@@ -727,9 +912,17 @@ int wpa_driver_wext_get_scan_results(void *priv,
 				       iwe->u.ap_addr.sa_data, ETH_ALEN);
 			}
 			break;
+		case SIOCGIWMODE:
+			if (ap_num >= max_size)
+				break;
+			if (iwe->u.mode == IW_MODE_ADHOC)
+				results[ap_num].caps |= IEEE80211_CAP_IBSS;
+			else if (iwe->u.mode == IW_MODE_MASTER ||
+				 iwe->u.mode == IW_MODE_INFRA)
+				results[ap_num].caps |= IEEE80211_CAP_ESS;
+			break;
 		case SIOCGIWESSID:
 			ssid_len = iwe->u.essid.length;
-			custom = pos + IW_EV_POINT_LEN;
 			if (custom + ssid_len > end)
 				break;
 			if (iwe->u.essid.flags &&
@@ -767,8 +960,9 @@ int wpa_driver_wext_get_scan_results(void *priv,
 			}
 			break;
 		case SIOCGIWENCODE:
-			if (ap_num < max_size)
-				results[ap_num].caps = iwe->u.data.flags;
+			if (ap_num < max_size &&
+			    !(iwe->u.data.flags & IW_ENCODE_DISABLED))
+				results[ap_num].caps |= IEEE80211_CAP_PRIVACY;
 			break;
 		case SIOCGIWRATE:
 			custom = pos + IW_EV_LCP_LEN;
@@ -791,7 +985,7 @@ int wpa_driver_wext_get_scan_results(void *priv,
 		case IWEVGENIE:
 			if (ap_num >= max_size)
 				break;
-			gpos = genie = pos + IW_EV_POINT_LEN;
+			gpos = genie = custom;
 			gend = genie + iwe->u.data.length;
 			if (gend > end) {
 				wpa_printf(MSG_INFO, "IWEVGENIE overflow");
@@ -824,7 +1018,6 @@ int wpa_driver_wext_get_scan_results(void *priv,
 			}
 			break;
 		case IWEVCUSTOM:
-			custom = pos + IW_EV_POINT_LEN;
 			clen = iwe->u.data.length;
 			if (custom + clen > end)
 				break;
@@ -892,32 +1085,64 @@ int wpa_driver_wext_get_scan_results(void *priv,
 static int wpa_driver_wext_get_range(void *priv)
 {
 	struct wpa_driver_wext_data *drv = priv;
-	struct iw_range range;
+	struct iw_range *range;
 	struct iwreq iwr;
 	int minlen;
+	size_t buflen;
+
+	/*
+	 * Use larger buffer than struct iw_range in order to allow the
+	 * structure to grow in the future.
+	 */
+	buflen = sizeof(struct iw_range) + 500;
+	range = malloc(buflen);
+	if (range == NULL)
+		return -1;
+	memset(range, 0, buflen);
 
 	memset(&iwr, 0, sizeof(iwr));
 	strncpy(iwr.ifr_name, drv->ifname, IFNAMSIZ);
-	iwr.u.data.pointer = (caddr_t) &range;
-	iwr.u.data.length = sizeof(range);
-	memset(&range, 0, sizeof(range));
+	iwr.u.data.pointer = (caddr_t) range;
+	iwr.u.data.length = buflen;
 
-	minlen = ((char *) &range.enc_capa) - (char *) &range +
-		sizeof(range.enc_capa);
+	minlen = ((char *) &range->enc_capa) - (char *) range +
+		sizeof(range->enc_capa);
 
 	if (ioctl(drv->ioctl_sock, SIOCGIWRANGE, &iwr) < 0) {
 		perror("ioctl[SIOCGIWRANGE]");
+		free(range);
 		return -1;
 	} else if (iwr.u.data.length >= minlen &&
-		   range.we_version_compiled >= 18) {
+		   range->we_version_compiled >= 18) {
 		wpa_printf(MSG_DEBUG, "SIOCGIWRANGE: WE(compiled)=%d "
 			   "WE(source)=%d enc_capa=0x%x",
-			   range.we_version_compiled, range.we_version_source,
-			   range.enc_capa);
-		/* TODO: Should report enc_capa to wpa_supplicant (i.e.,
-		 * whether the driver supports WPA/WPA2 and TKIP/CCMP). */
+			   range->we_version_compiled,
+			   range->we_version_source,
+			   range->enc_capa);
+		drv->has_capability = 1;
+		drv->we_version_compiled = range->we_version_compiled;
+		if (range->enc_capa & IW_ENC_CAPA_WPA) {
+			drv->capa.key_mgmt |= WPA_DRIVER_CAPA_KEY_MGMT_WPA |
+				WPA_DRIVER_CAPA_KEY_MGMT_WPA_PSK;
+		}
+		if (range->enc_capa & IW_ENC_CAPA_WPA2) {
+			drv->capa.key_mgmt |= WPA_DRIVER_CAPA_KEY_MGMT_WPA2 |
+				WPA_DRIVER_CAPA_KEY_MGMT_WPA2_PSK;
+		}
+		drv->capa.enc |= WPA_DRIVER_CAPA_ENC_WEP40 |
+			WPA_DRIVER_CAPA_ENC_WEP104;
+		if (range->enc_capa & IW_ENC_CAPA_CIPHER_TKIP)
+			drv->capa.enc |= WPA_DRIVER_CAPA_ENC_TKIP;
+		if (range->enc_capa & IW_ENC_CAPA_CIPHER_CCMP)
+			drv->capa.enc |= WPA_DRIVER_CAPA_ENC_CCMP;
+		wpa_printf(MSG_DEBUG, "  capabilities: key_mgmt 0x%x enc 0x%x",
+			   drv->capa.key_mgmt, drv->capa.enc);
+	} else {
+		wpa_printf(MSG_DEBUG, "SIOCGIWRANGE: too old (short) data - "
+			   "assuming WPA is not supported");
 	}
 
+	free(range);
 	return 0;
 }
 
@@ -927,9 +1152,6 @@ static int wpa_driver_wext_set_wpa(void *priv, int enabled)
 	struct wpa_driver_wext_data *drv = priv;
 	wpa_printf(MSG_DEBUG, "%s", __FUNCTION__);
 
-	if (enabled)
-		wpa_driver_wext_get_range(drv);
-
 	return wpa_driver_wext_set_auth_param(drv, IW_AUTH_WPA_ENABLED,
 					      enabled);
 }
@@ -1005,8 +1227,15 @@ static int wpa_driver_wext_set_key_ext(void *priv, wpa_alg alg,
 	}
 
 	if (ioctl(drv->ioctl_sock, SIOCSIWENCODEEXT, &iwr) < 0) {
+		ret = errno == EOPNOTSUPP ? -2 : -1;
+		if (errno == ENODEV) {
+			/*
+			 * ndiswrapper seems to be returning incorrect error
+			 * code.. */
+			ret = -2;
+		}
+
 		perror("ioctl[SIOCSIWENCODEEXT]");
-		ret = -1;
 	}
 
 	free(ext);
@@ -1014,10 +1243,10 @@ static int wpa_driver_wext_set_key_ext(void *priv, wpa_alg alg,
 }
 
 
-static int wpa_driver_wext_set_key(void *priv, wpa_alg alg,
-				   const u8 *addr, int key_idx,
-				   int set_tx, const u8 *seq, size_t seq_len,
-				   const u8 *key, size_t key_len)
+int wpa_driver_wext_set_key(void *priv, wpa_alg alg,
+			    const u8 *addr, int key_idx,
+			    int set_tx, const u8 *seq, size_t seq_len,
+			    const u8 *key, size_t key_len)
 {
 	struct wpa_driver_wext_data *drv = priv;
 	struct iwreq iwr;
@@ -1033,7 +1262,7 @@ static int wpa_driver_wext_set_key(void *priv, wpa_alg alg,
 	if (ret == 0)
 		return 0;
 
-	if (ret == -EOPNOTSUPP &&
+	if (ret == -2 &&
 	    (alg == WPA_ALG_NONE || alg == WPA_ALG_WEP)) {
 		wpa_printf(MSG_DEBUG, "Driver did not support "
 			   "SIOCSIWENCODEEXT, trying SIOCSIWENCODE");
@@ -1203,21 +1432,18 @@ wpa_driver_wext_associate(void *priv,
 
 	wpa_printf(MSG_DEBUG, "%s", __FUNCTION__);
 
-	if (wpa_driver_wext_set_mode(drv, params->mode) < 0)
-		ret = -1;
-	if (wpa_driver_wext_set_gen_ie(drv, params->wpa_ie, params->wpa_ie_len)
-	    < 0)
-		ret = -1;
-	if (wpa_driver_wext_set_freq(drv, params->freq) < 0)
-		ret = -1;
-	if (wpa_driver_wext_set_ssid(drv, params->ssid, params->ssid_len) < 0)
-		ret = -1;
-	if (wpa_driver_wext_set_bssid(drv, params->bssid) < 0)
+	if (!params->bssid &&
+	    wpa_driver_wext_set_bssid(drv, NULL) < 0)
 		ret = -1;
 
+	if (wpa_driver_wext_set_mode(drv, params->mode) < 0)
+		ret = -1;
 	/* TODO: should consider getting wpa version and cipher/key_mgmt suites
 	 * from configuration, not from here, where only the selected suite is
 	 * available */
+	if (wpa_driver_wext_set_gen_ie(drv, params->wpa_ie, params->wpa_ie_len)
+	    < 0)
+		ret = -1;
 	if (params->wpa_ie == NULL || params->wpa_ie_len == 0)
 		value = IW_AUTH_WPA_VERSION_DISABLED;
 	else if (params->wpa_ie[0] == RSN_INFO_ELEM)
@@ -1239,6 +1465,13 @@ wpa_driver_wext_associate(void *priv,
 	if (wpa_driver_wext_set_auth_param(drv,
 					   IW_AUTH_KEY_MGMT, value) < 0)
 		ret = -1;
+	value = params->key_mgmt_suite != KEY_MGMT_NONE ||
+		params->pairwise_suite != CIPHER_NONE ||
+		params->group_suite != CIPHER_NONE ||
+		params->wpa_ie_len;
+	if (wpa_driver_wext_set_auth_param(drv,
+					   IW_AUTH_PRIVACY_INVOKED, value) < 0)
+		ret = -1;
 
 	/* Allow unencrypted EAPOL messages even if pairwise keys are set when
 	 * not using WPA. IEEE 802.1X specifies that these frames are not
@@ -1253,6 +1486,13 @@ wpa_driver_wext_associate(void *priv,
 					   IW_AUTH_RX_UNENCRYPTED_EAPOL,
 					   allow_unencrypted_eapol) < 0)
 		ret = -1;
+	if (params->freq && wpa_driver_wext_set_freq(drv, params->freq) < 0)
+		ret = -1;
+	if (wpa_driver_wext_set_ssid(drv, params->ssid, params->ssid_len) < 0)
+		ret = -1;
+	if (params->bssid &&
+	    wpa_driver_wext_set_bssid(drv, params->bssid) < 0)
+		ret = -1;
 
 	return ret;
 }
@@ -1298,7 +1538,68 @@ int wpa_driver_wext_set_mode(void *priv, int mode)
 }
 
 
-struct wpa_driver_ops wpa_driver_wext_ops = {
+static int wpa_driver_wext_pmksa(struct wpa_driver_wext_data *drv,
+				 u32 cmd, const u8 *bssid, const u8 *pmkid)
+{
+	struct iwreq iwr;
+	struct iw_pmksa pmksa;
+	int ret = 0;
+
+	memset(&iwr, 0, sizeof(iwr));
+	strncpy(iwr.ifr_name, drv->ifname, IFNAMSIZ);
+	memset(&pmksa, 0, sizeof(pmksa));
+	pmksa.cmd = cmd;
+	pmksa.bssid.sa_family = ARPHRD_ETHER;
+	if (bssid)
+		memcpy(pmksa.bssid.sa_data, bssid, ETH_ALEN);
+	if (pmkid)
+		memcpy(pmksa.pmkid, pmkid, IW_PMKID_LEN);
+	iwr.u.data.pointer = (caddr_t) &pmksa;
+	iwr.u.data.length = sizeof(pmksa);
+
+	if (ioctl(drv->ioctl_sock, SIOCSIWPMKSA, &iwr) < 0) {
+		perror("ioctl[SIOCSIWPMKSA]");
+		ret = -1;
+	}
+
+	return ret;
+}
+
+
+static int wpa_driver_wext_add_pmkid(void *priv, const u8 *bssid,
+				     const u8 *pmkid)
+{
+	struct wpa_driver_wext_data *drv = priv;
+	return wpa_driver_wext_pmksa(drv, IW_PMKSA_ADD, bssid, pmkid);
+}
+
+
+static int wpa_driver_wext_remove_pmkid(void *priv, const u8 *bssid,
+		 			const u8 *pmkid)
+{
+	struct wpa_driver_wext_data *drv = priv;
+	return wpa_driver_wext_pmksa(drv, IW_PMKSA_REMOVE, bssid, pmkid);
+}
+
+
+static int wpa_driver_wext_flush_pmkid(void *priv)
+{
+	struct wpa_driver_wext_data *drv = priv;
+	return wpa_driver_wext_pmksa(drv, IW_PMKSA_FLUSH, NULL, NULL);
+}
+
+
+static int wpa_driver_wext_get_capa(void *priv, struct wpa_driver_capa *capa)
+{
+	struct wpa_driver_wext_data *drv = priv;
+	if (!drv->has_capability)
+		return -1;
+	memcpy(capa, &drv->capa, sizeof(*capa));
+	return 0;
+}
+
+
+const struct wpa_driver_ops wpa_driver_wext_ops = {
 	.name = "wext",
 	.desc = "Linux wireless extensions (generic)",
 	.get_bssid = wpa_driver_wext_get_bssid,
@@ -1315,4 +1616,8 @@ struct wpa_driver_ops wpa_driver_wext_ops = {
 	.set_auth_alg = wpa_driver_wext_set_auth_alg,
 	.init = wpa_driver_wext_init,
 	.deinit = wpa_driver_wext_deinit,
+	.add_pmkid = wpa_driver_wext_add_pmkid,
+	.remove_pmkid = wpa_driver_wext_remove_pmkid,
+	.flush_pmkid = wpa_driver_wext_flush_pmkid,
+	.get_capa = wpa_driver_wext_get_capa,
 };
