@@ -1678,6 +1678,14 @@ static void wpa_supplicant_process_1_of_2(struct wpa_supplicant *wpa_s,
 					   "AES-WRAP len %d", keydatalen);
 				return;
 			}
+			if (maxkeylen > sizeof(gtk)) {
+				wpa_printf(MSG_WARNING, "WPA: AES-WRAP key "
+					   "data too long (keydatalen=%lu "
+					   "maxkeylen=%lu)",
+					   (unsigned long) keydatalen,
+					   (unsigned long) maxkeylen);
+				return;
+			}
 			if (aes_unwrap(wpa_s->ptk.encr_key, maxkeylen / 8,
 				       (u8 *) (key + 1), gtk)) {
 				wpa_printf(MSG_WARNING, "WPA: AES unwrap "
